from random import randint
from typing import List, Optional

from rolling_dice.models.errors.DiceError import DiceError


class Dice:
    __slots__ = ("_throw",
                 "_face",
                 "_retain_f",
                 "_retain_n",
                 "_result",
                 "_retain",
                 )

    def __init__(self, throw,
                 face,
                 retain_f=None,
                 retain_n=None) -> None:
        self._throw = throw
        self._face = face
        self._retain_f = retain_f
        self._retain_n = retain_n

        self._result = None
        self._retain = None

    def to_str(self, view_retains=False, startswith_retain='<strike>', endswith_retain='</strike>') -> str:
        if (not isinstance(self._all_result, list) or not self.retains) and view_retains:
            view_retains = False
        if view_retains and isinstance(self._all_result, list) and len(self.result) != len(self._all_result):
            values = list(map(str, self._all_result)) if isinstance(self._all_result, list) else [str(self._all_result)]
            retains, cut = list(map(str, self.retains)), []
            for n, value in enumerate(values):
                if value in retains:
                    retains.remove(value)
                else:
                    cut.append(value)
                    values[n] = f"{startswith_retain}{value}{endswith_retain}"
        else:
            values = map(str, self.result) if isinstance(self.result, list) else [str(self.result)]

        return f"[{'+'.join(values)}]"

    def __repr__(self) -> str:
        return f"Dice(throw={self.throw}, face={self.face}, retain_f = {self._retain_f}, " \
               f"retain_n = {self._retain_n}, result = {self.result})"

    def _get_retains(self) -> Optional[List[int]]:
        if self._retain_n is None and self._retain_f is None:
            return None
        elif not all((self._retain_f, self._retain_n)):
            raise DiceError("Unspecified retain formula or count for save")
        elif self._retain_n > self._throw:
            return None
        all_results = self._all_result.copy()
        results = []

        for _ in range(self._retain_n):
            exclude = self._retain_f(all_results)
            all_results.remove(exclude)
            results.append(exclude)
        return results

    @property
    def _all_result(self) -> List[int]:
        if not self._result:
            self._result = [randint(1, self._face) for _ in range(self._throw)]
        return self._result

    @property
    def throw(self) -> int:
        return self._throw

    @property
    def face(self) -> int:
        return self._face

    @property
    def retains(self) -> Optional[List[int]]:
        if self._retain is None:
            self._retain = self._get_retains()
        return self._retain

    @property
    def result(self) -> Optional[List[int]]:
        list_results = self._all_result.copy()
        if self.retains:
            list_results = self._retain
        return list_results

    @property
    def total(self) -> int:
        return sum(self.result)



from re import search, findall
from typing import List

from rolling_dice.models.Dice import Dice
from rolling_dice.models.Result import Result
from rolling_dice.models.errors.ParseError import ParseError
from rolling_dice.parser.lark_laboers import open_lark
from rolling_dice.resources import GRAMMAR_DICE, GRAMMAR_CALCULATOR


def get_result(text,
               grammar_dice=GRAMMAR_DICE,
               grammar_calc=GRAMMAR_CALCULATOR) -> List:
    results = []
    repeats_math = search(r"(^\d+)[хx]|[хx](\d+$)", text)
    repeats = repeats_math.group(1) or repeats_math.group(2) if repeats_math else 1
    for _ in range(int(repeats) if int(repeats) < 10 else 10):
        t = text.replace(repeats_math.group(0), "") if repeats_math else text
        result = Result(raw=t)
        result.dices = []
        for dice in findall(r"(\d*[dkдк]\d+[hlвнd]?\d*)", t):
            value: Dice = open_lark(text=dice, grammar=grammar_dice)
            result.dices.append((dice, value))
        result.total = open_lark(text=result.replaced_dices, grammar=grammar_calc)
        if str(result.total) == t:
            raise ParseError
        results.append(result)
    return results



from dataclasses import dataclass
from typing import Tuple, List

from rolling_dice.models.Dice import Dice


@dataclass
class Result:
    raw: str
    total: int = None
    dices: List[Tuple[str, Dice]] = None

    @property
    def total_formula(self) -> str:
        result = self.raw
        for dice, cls in self.dices:
            result = result.replace(dice, cls.to_str(view_retains=True, ), 1)
        return result + f"={self.total}"

    @property
    def replaced_dices(self) -> str:
        result = self.raw
        for dice, cls in self.dices:
            result = result.replace(dice, str(cls.total), 1)
        return result


import sys

if sys.version_info < (3, 7):
    raise ImportError('Your Python version {0} is not supported by aiogram, please install '
                      'Python 3.7+'.format('.'.join(map(str, sys.version_info[:3]))))

from rolling_dice.models import Dice, Result, DiceError, NotFoundMethod, ParseError
from rolling_dice.parser import get_result, open_lark
from rolling_dice.resources import GRAMMAR_CALCULATOR, GRAMMAR_DICE

__all__ = (
    "Dice",
    "Result",
    "DiceError",
    "NotFoundMethod",
    "ParseError",
    "get_result",
    "open_lark",
    "GRAMMAR_CALCULATOR",
    "GRAMMAR_DICE"
)

__version__ = '1.1'



from rolling_dice.models.errors.DiceError import DiceError
from rolling_dice.models.errors.NotFoundMethod import NotFoundMethod
from rolling_dice.models.errors.ParseError import ParseError


__all__ = (
    "DiceError",
    "NotFoundMethod",
    "ParseError"
)


from operator import add, mul, sub, truediv as div
from typing import Union

from lark import Tree, Lark

from rolling_dice.models.Dice import Dice


def simple_calculation(tree: Tree) -> int:
    values = [get_next_point(child) for child in tree.children]
    if tree.data == "add":
        return add(*values)
    elif tree.data == "sub":
        return sub(*values)
    elif tree.data == "mul":
        return mul(*values)
    elif tree.data == "div":
        return div(*values)


def parse_roll_dice(tree: Tree) -> Dice:
    if len(tree.children) == 2:
        thrown, face = [get_next_point(child) for child in tree.children]
    else:
        thrown, face = 1, get_next_point(*tree.children)
    return Dice(throw=thrown, face=face)


def filtration_dices(tree: Tree) -> Dice:
    dice: Dice = get_next_point(tree.children[0])

    if tree.data == "max":
        dice._retain_f = max
    elif tree.data == "min":
        dice._retain_f = min
    dice._retain_n = get_next_point(tree.children[1]) if len(tree.children) > 1 else 1
    return dice


def get_next_point(tree: Tree) -> Union[int, Dice, Tree]:
    if tree.data in ("add", "sub", "mul", "div"):
        return simple_calculation(tree)
    elif tree.data == "to_int":
        return int(tree.children[0])
    elif tree.data == "res":
        return sum([get_next_point(child) for child in tree.children])
    elif tree.data == "dice":
        return parse_roll_dice(tree)
    elif tree.data in ("max", "min"):
        return filtration_dices(tree)


def open_lark(text, grammar):
    return get_next_point(Lark(grammar, start="start").parse(text))


from rolling_dice.models.errors import DiceError, NotFoundMethod, ParseError
from rolling_dice.models.Dice import Dice
from rolling_dice.models.Result import Result

__all__ = (
    "Dice",
    "Result",
    "DiceError",
    "NotFoundMethod",
    "ParseError"
)



#!/usr/bin/env python3
import pathlib
import re
import sys

from setuptools import find_packages, setup

WORK_DIR = pathlib.Path(__file__).parent
MINIMAL_PY_VERSION = (3, 7)

if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('rolling_dice works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))


def get_version() -> str:
    """
    Read version

    :return: str
    """
    txt = (WORK_DIR / 'rolling_dice' / '__init__.py').read_text('utf-8')
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


setup(
    name='rolling_dice',
    version=get_version(),
    packages=find_packages(exclude=('tests', 'tests.*')),
    url='https://github.com/M9SCO/diceroll',
    license='MIT',
    author='Chernov M_9SCO Grigorii',
    author_email='chr.grigorii@gmail.com',
    python_requires='>=3.7',
    description='Is a pretty simple library for rolling dice',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    install_requires=[
        'lark>=1.1.2'
    ],
    include_package_data=False,
)



from rolling_dice.parser.lark_laboers import (
    simple_calculation,
    parse_roll_dice,
    filtration_dices,
    get_next_point,
    open_lark)
from rolling_dice.parser.get_result import get_result

__all__ = (
    "simple_calculation",
    "parse_roll_dice",
    "filtration_dices",
    "get_next_point",
    "open_lark",
    "get_result",
)



GRAMMAR_CALCULATOR = """\
?start      : s_math

?s_math     : s_math "+" s_math         -> add
            | s_math "-" s_math         -> sub
            | h_math

?h_math     : h_math "*" h_math         -> mul
            | h_math "/" h_math         -> div
            | result
            | num

?result     : num                -> res

?num        : "(" s_math ")"
            | "[" s_math "]"
            | NUMBER                    -> to_int

%import common.NUMBER
%ignore " "
"""

GRAMMAR_DICE = """\
?start      : filtration
?filtration : filtration "h" num? -> max
            | filtration "в" num? -> max
            | filtration "l" num? -> min
            | filtration "н" num? -> min
            | dice
?dice       : num? "d" num        -> dice
            | num? "к" num        -> dice
            | num? "k" num        -> dice
            | num? "д" num        -> dice
?num        : NUMBER              -> to_int
%import common.NUMBER
%ignore " "
"""